﻿using System.Collections.Immutable;
using Connected.Entities;
using Connected.Entities.Storage;
using Connected.Notifications.Events;
using Connected.Security.Permissions;
using Connected.ServiceModel;
using Connected.Services;

namespace Connected.Common.Security.Permissions;

internal sealed class DeletePermission : ServiceAction<PrimaryKeyArgs<int>>
{
	public DeletePermission(IPermissionService permissionService, IStorageProvider storage, IPermissionCache cache, IEventService events)
	{
		PermissionService = permissionService;
		Storage = storage;
		Cache = cache;
		Events = events;
	}

	private IPermissionService PermissionService { get; }
	private IStorageProvider Storage { get; }
	private IPermissionCache Cache { get; }
	private IEventService Events { get; }

	protected override async Task OnInvoke()
	{
		await Storage.Open<Permission>().Update(new Permission { Id = Arguments.Id, State = State.Deleted });
	}

	protected override async Task OnCommitted()
	{
		await Cache.Refresh(Arguments.Id);
		await Events.Enqueue(this, PermissionService, ServiceEvents.Deleted, Arguments.Id);
	}
}

internal sealed class QueryPermissions : ServiceFunction<IDto, ImmutableList<IPermission>?>
{
	public QueryPermissions(IPermissionCache permissionCache)
	{
		PermissionCache = permissionCache;
	}

	private IPermissionCache PermissionCache { get; }
	protected override async Task<ImmutableList<IPermission>?> OnInvoke()
	{
		return await (from dc in PermissionCache
					  select dc).AsEntities<IPermission>();
	}
}

internal sealed class SearchPermissions : ServiceFunction<PermissionSearchArgs, ImmutableList<IPermission>?>
{
	public SearchPermissions(IPermissionCache permissionCache)
	{
		PermissionCache = permissionCache;
	}

	private IPermissionCache PermissionCache { get; }
	protected override async Task<ImmutableList<IPermission>?> OnInvoke()
	{
		return await (from dc in PermissionCache
					  where (string.IsNullOrEmpty(Arguments.Entity) || string.Equals(Arguments.Entity, dc.Entity, StringComparison.OrdinalIgnoreCase))
					  && (string.IsNullOrEmpty(Arguments.Claim) || string.Equals(Arguments.Claim, dc.Claim, StringComparison.OrdinalIgnoreCase))
					  && (string.IsNullOrEmpty(Arguments.PrimaryKey) || string.Equals(Arguments.PrimaryKey, dc.PrimaryKey, StringComparison.OrdinalIgnoreCase))
					  select dc).AsEntities<IPermission>();
	}
}

internal sealed class SelectPermission : ServiceFunction<PrimaryKeyArgs<int>, IPermission?>
{
	public SelectPermission(IPermissionCache permissionCache)
	{
		PermissionCache = permissionCache;
	}

	private IPermissionCache PermissionCache { get; }
	protected override async Task<IPermission?> OnInvoke()
	{
		return await (from dc in PermissionCache
					  where dc.Id == Arguments.Id
					  select dc).AsEntity();
	}
}

internal sealed class InsertPermission : ServiceFunction<PermissionArgs, int>
{
	public InsertPermission(IPermissionService permissionService, IStorageProvider storage, IEventService events, IPermissionCache cache)
	{
		PermissionService = permissionService;
		Storage = storage;
		Events = events;
		Cache = cache;
	}

	private IPermissionService PermissionService { get; }
	private IStorageProvider Storage { get; }
	private IEventService Events { get; }
	private IPermissionCache Cache { get; }

	protected override async Task<int> OnInvoke()
	{
		var entity = Arguments.AsEntity<Permission>(State.New);
		var result = await Storage.Open<Permission>().Update(entity);

		return result.Id;
	}

	protected override async Task OnCommitted()
	{
		await Cache.Refresh(Result);
		await Events.Enqueue(this, PermissionService, ServiceEvents.Inserted, Result);
	}
}

internal sealed class UpdatePermission : ServiceAction<PermissionUpdateArgs>
{
	public UpdatePermission(IPermissionService permissionService, IStorageProvider storage, IPermissionCache cache, IEventService events)
	{
		PermissionService = permissionService;
		Storage = storage;
		Cache = cache;
		Events = events;
	}

	private IPermissionService PermissionService { get; }
	private IStorageProvider Storage { get; }
	private IPermissionCache Cache { get; }
	private IEventService Events { get; }

	protected override async Task OnInvoke()
	{
		await Storage.Open<Permission>().Update(await Load(), Arguments, async () =>
		{
			await Cache.Refresh(Arguments.Id);

			return await Load();
		});
	}

	private async Task<Permission?> Load() => await (from dc in Cache where dc.Id == Arguments.Id select dc).AsEntity();

	protected override async Task OnCommitted()
	{
		await Cache.Refresh(Arguments.Id);
		await Events.Enqueue(this, PermissionService, ServiceEvents.Updated, Arguments.Id);
	}
}